/*
 * gpio_etwarm.c
 *
 * This programm demonstates the access to GPIO from user space
 * with the mmap systemcall. 
 * This sample program bases on http://www.simtec.co.uk/appnotes/AN0014/
 * 
 * (C) Copyright 2007
 * Forschungs- und Transferzentrum (FTZ) Leipzig
 * Carsten Koegler <koegler@ftz-leipzig.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <sys/mman.h>
#include "AT91RM9200.h"
#include <stdio.h>
#include <fcntl.h>		// file access
#include <unistd.h>		// needed for close()
#include "gpio_etwarm.h"

#define MAP_SIZE 4096UL			/* mmap() uses this page size */
#define MAP_MASK (MAP_SIZE - 1) /* bitmask to calculate page start address */

int 	  fd_gpio_etwarm;
void	  *map_base_gpio_etwarm;
AT91PS_PMC p_pms_struct;
AT91PS_PIO p_pioa_struct;
AT91PS_PIO p_piob_struct;
AT91PS_PIO p_pioc_struct;
AT91PS_PIO p_piod_struct;


// we have to access these hardware addresses
// these addresses are all within one 4kByte page
#define AT91C_BASE_PIOA      ((AT91PS_PIO) 	0xFFFFF400) // (PIOA) Base Address
/*
#define AT91C_BASE_PIOB      ((AT91PS_PIO) 	0xFFFFF600) // (PIOB) Base Address
#define AT91C_BASE_PIOC      ((AT91PS_PIO) 	0xFFFFF800) // (PIOC) Base Address
#define AT91C_BASE_PIOD      ((AT91PS_PIO) 	0xFFFFFA00) // (PIOD) Base Address
#define AT91_PMC_SCER_ADDRESS 0xFFFFFC00 // Power Management Controller (System Clock Enable Register)
*/


/*!****************************************************************************
 * \brief		Initializes interface to GPIOs
 *
 ******************************************************************************
 */
void gpio_etwarm_init(void)
{

	// to use an PIO we map a page of /dev/mem in our virtual process memory
    if ((fd_gpio_etwarm = open("/dev/mem", O_RDWR | O_SYNC)) == -1) {
        printf("GPIO-ETWARM: /dev/mem could not be opened.\n");
    } 
    /* Map physical memory to a logical address */
    map_base_gpio_etwarm = mmap(0, 			/* We don't specify target address */
					MAP_SIZE, 	/* we want to access 1 page */
					PROT_READ | PROT_WRITE, MAP_SHARED, /* memory protection flags */
					fd_gpio_etwarm,	/* file descriptor of /dev/mem */
					(int)AT91C_BASE_PIOA & ~MAP_MASK);	/* physical address (begin of a page) */
    if (map_base_gpio_etwarm == (void *) -1) {
        printf("GPIO-ETWARM: Memory map failed.\n");
    } 
/*
    else {
        printf("GPIO-ETWARM: Memory mapped at address 0x%08X.\n", (int)map_base);
	}
*/

    /* calculate virtual address for registers that we need */
	p_pioa_struct = map_base_gpio_etwarm + ((int)AT91C_BASE_PIOA & MAP_MASK);
	p_piob_struct = (AT91PS_PIO)((int)p_pioa_struct + 0x200);
	p_pioc_struct = (AT91PS_PIO)((int)p_pioa_struct + 0x400);
	p_piod_struct = (AT91PS_PIO)((int)p_pioa_struct + 0x800);
	p_pms_struct = (AT91PS_PMC)((int)p_pioa_struct + 0xA00);
}


/*!****************************************************************************
 * \brief		Configs GPIO port pins
 *
 * \note		gpio_etwarm_init() MUST be called before.
 * @param[in]	port    - GPIO port range: AT91_PORTA .. AT91_PORTD
 * @param[in]	bitmask - 32 Bit bitmask of Portpins to manipulate
 * @param[out]	cfg  	- specifies in/out, see T_AT91_PORT_CFG
 ******************************************************************************
 */
void gpio_etwarm_cfg(T_AT91_PORT_NO port, int bitmask, T_AT91_PORT_CFG cfg)
{

	/* enable periphal clock in the power management controller*/

//	printf("GPIO-ETWARM: PMC_PCSR before: 0x%08X\n", p_pms_struct->PMC_PCSR);
	switch (port) {
		case AT91_PORTA:
			p_pms_struct->PMC_PCER = (1<<AT91C_ID_PIOA);
			p_pioa_struct->PIO_PER = bitmask;
			if (cfg == AT91_PORT_OUTPUT)
				p_pioa_struct->PIO_OER = bitmask;
			else
				p_pioa_struct->PIO_ODR = bitmask;
			break;
		case AT91_PORTB:
			p_pms_struct->PMC_PCER = (1<<AT91C_ID_PIOB);
			p_piob_struct->PIO_PER = bitmask;
			if (cfg == AT91_PORT_OUTPUT)
				p_piob_struct->PIO_OER = bitmask;
			else
				p_piob_struct->PIO_ODR = bitmask;
			break;
		case AT91_PORTC:
			p_pms_struct->PMC_PCER = (1<<AT91C_ID_PIOC);
			p_pioc_struct->PIO_PER = bitmask;
			if (cfg == AT91_PORT_OUTPUT)
				p_pioc_struct->PIO_OER = bitmask;
			else
				p_pioc_struct->PIO_ODR = bitmask;
			break;
		case AT91_PORTD:
			p_pms_struct->PMC_PCER = (1<<AT91C_ID_PIOD);
			p_piod_struct->PIO_PER = bitmask;
			if (cfg == AT91_PORT_OUTPUT)
				p_pioc_struct->PIO_OER = bitmask;
			else
				p_pioc_struct->PIO_ODR = bitmask;
			break;
	}
//	printf("GPIO-ETWARM: PMC_PCSR after: 0x%08X\n", p_pms_struct->PMC_PCSR);
}


/*!****************************************************************************
 * \brief		Set GPIO pins to given value.
 *
 * @param[in]	port    - GPIO port range: AT91_PORTA .. AT91_PORTD
 * @param[in]	bitmask - 32 Bit bitmask of Portpins to manipulate
 * @param[out]	value   - 32 Bit bitmask with new value
 ******************************************************************************
 */
void gpio_etwarm_out(T_AT91_PORT_NO port, int bitmask, int value)
{
	switch (port) {
		case AT91_PORTA:
			p_pioa_struct->PIO_SODR = bitmask & value;
			value = ~value;
			p_pioa_struct->PIO_CODR = bitmask & value;
			break;
		case AT91_PORTB:
			p_piob_struct->PIO_SODR = bitmask & value;
			value = ~value;
			p_piob_struct->PIO_CODR = bitmask & value;
			break;
		case AT91_PORTC:
			p_pioc_struct->PIO_SODR = bitmask & value;
			value = ~value;
			p_pioc_struct->PIO_CODR = bitmask & value;
			break;
		case AT91_PORTD:
			p_piod_struct->PIO_SODR = bitmask & value;
			value = ~value;
			p_piod_struct->PIO_CODR = bitmask & value;
			break;
	}
}


/*!****************************************************************************
 * \brief		Reads state of GPIO pins
 *
 * @param[in]	port    - GPIO port range: AT91_PORTA .. AT91_PORTD
 * @param[in]	bitmask - 32 Bit bitmask of Portpins to read
 * @return		OR-ed state of all specified port pins
 ******************************************************************************
 */
int gpio_etwarm_in(T_AT91_PORT_NO port, int bitmask)
{
	switch (port) {
		case AT91_PORTA:
			return (p_pioa_struct->PIO_PDSR&bitmask) != 0x0;
		case AT91_PORTB:
			return (p_piob_struct->PIO_PDSR&bitmask) != 0x0;
		case AT91_PORTC:
			return (p_pioc_struct->PIO_PDSR&bitmask) != 0x0;
		case AT91_PORTD:
			return (p_piod_struct->PIO_PDSR&bitmask) != 0x0;
	}
return 0;
}


/*!****************************************************************************
 * \brief		Exit function
 *
 * \note		We do NOT revert changes to the GPIO registers!
 ******************************************************************************
 */
void gpio_etwarm_exit(void)
{
	/* unmapping the physical memory */
	if (munmap(map_base_gpio_etwarm, MAP_SIZE) == -1)
		printf("GPIO-ETWARM: Memory unmap failed.\n");

	close(fd_gpio_etwarm);
}
